#include <QProcess>
#include <QStringList>
#include <QUrl>

#include <QDebug>

#include "logger.h"
#include "ui/widgets/browser/browser.h"
#include "ui/widgets/document.h"
#include "ui/widgets/idle.h"
#include "ui/widgets/image.h"
#include "ui/widgets/pdf.h"
#include "ui/widgets/video.h"
#include "ui/widgets/webpage.h"
#include "playlistparser.h"
#include "downloader.h"
#include "dispatcher.h"

Dispatcher::Dispatcher(Viewer *viewer, QString path)
    : QObject()
    , m_status(Dispatcher::StatusNone)
    , m_currentWidget(0)
    , m_prevWidget(0)
    , m_viewer(viewer)
    , m_plRefresh(false)
    , m_idling(false)
{
    settings = Settings::load();
    m_mode = settings->get("main/mode").toString();

    connect(viewer, SIGNAL(end()), this, SLOT(next()));
    connect(viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));

    if (settings->get("main/no_screensaver").toBool())
        disableScreensaver();
}

void Dispatcher::StartSession()
{
    if ((m_status == Dispatcher::StatusBrowser) ||
            (m_status == Dispatcher::StatusAfterBrowser))
        return;

    m_currentWidget->stop();
    startBrowser();
}

void Dispatcher::StopSession()
{
    if (m_status != Dispatcher::StatusBrowser)
        return;

    m_currentWidget->stop();
    startSlideshow();
}

void Dispatcher::AddTime(int seconds)
{
    // TODO
}

void Dispatcher::onScreensaverTimeout()
{
    QProcess p;
    QStringList args;
    args << "-deactivate";
    p.start("xscreensaver-command", args);
    p.waitForFinished();
}

void Dispatcher::disableScreensaver()
{
    connect(&m_ssTimer, SIGNAL(timeout()), this, SLOT(onScreensaverTimeout()));
    m_ssTimer.start(30000);
}

void Dispatcher::onPlTimeout()
{
    m_plRefresh = true;
}

void Dispatcher::startup()
{
    connect(&m_plTimer, SIGNAL(timeout()), this, SLOT(onPlTimeout()));
    int timeout = settings->get("slideshow/refresh_after_minutes").toInt();
    m_plTimer.start(60000 * timeout);

    if ((m_mode == "kiosk") &&
            settings->get("main/start_with_session").toBool())
        startBrowser();
    else
        startSlideshow();
}

void Dispatcher::onMouseEvent()
{
    if (m_mode != "kiosk")
        return;

    m_currentWidget->stop();
    m_loopIdx = -1;
    startBrowser();
}

void Dispatcher::onReconnectMouseEventTimeout()
{
    connect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
}

void Dispatcher::reconnectMouseEvent()
{
    QTimer::singleShot(1000, this, SLOT(onReconnectMouseEventTimeout()));
}

void Dispatcher::next()
{
    if (m_status == Dispatcher::StatusNone)
    {
        startup();
    }

    else if (m_status == Dispatcher::StatusBrowser)
    {
        qDebug() << "dispatcher: End StatusBrowser";
        disconnect(m_viewer, SIGNAL(mouseEvent()), this, SLOT(onMouseEvent()));
        execHook("session-stop");
        afterBrowser();
    }

    else if (m_status == Dispatcher::StatusAfterBrowser)
    {
        qDebug() << "dispatcher: End StatusAfterBrowser";
        if (settings->get("main/after_web_session").toString() == "session")
        {
            Logger::logInfo("STARTING: web session");
            startBrowser();
        }
        else
        {
            reconnectMouseEvent();
            startSlideshow();
        }
    }

    else
    {
        showLoopNext();
    }
}

void Dispatcher::afterBrowser()
{
    m_status = Dispatcher::StatusAfterBrowser;
    QString uri = settings->get("main/post_session_uri").toString();
    if (uri.isEmpty())
    {
        next();
        return;
    }
    int time = settings->get("main/post_session_uri_time").toInt();
    PlaylistItem item(uri, PlaylistItem::TypeWebpage, time);
    m_currentWidget = new Webpage(&item, m_viewer);
    m_viewer->next(dynamic_cast<QWidget *>(m_currentWidget));
}

void Dispatcher::startBrowser()
{
    execHook("session-start");
    Logger::logInfo("STARTING: web session");
    m_status = Dispatcher::StatusBrowser;
    m_currentWidget = new Browser;
    m_viewer->next(dynamic_cast<QWidget *>(m_currentWidget));
}

void Dispatcher::execHook(QString hook)
{
    QString hookFile = QString("/etc/xdg/webkiosk/hooks/%1").arg(hook);
    QProcess p;
    p.start(hookFile);
    p.waitForFinished();
}

void Dispatcher::startSlideshow()
{
    QUrl url;

    if (settings->get("slideshow/use_remote").toBool())
        url = QUrl(settings->get("slideshow/pages").toString());
    else
        url = QUrl(QString("file://%1").arg(settings->playlistFile()));

    m_pl = PlaylistParser::get(url);
    m_loopIdx = -1;
    m_loop = m_pl->selectLoop();
    if (m_loop->downloadable())
    {
        Downloader *d = new Downloader(m_loop);
        connect(d, SIGNAL(downloadDone()), this, SLOT(showLoopNext()));
        d->start();
    }
    else
    {
        showLoopNext();
    }
}

void Dispatcher::showLoopNext()
{
    m_status = Dispatcher::StatusPlaylist;

    ++m_loopIdx;
    if (m_loopIdx >= m_loop->count())
    {
        /* We need to download the playlist to check what's changed */
        if (m_plRefresh)
        {
            m_plRefresh = false;
            startSlideshow();
            return;
        }

        m_pl->refresh();
        m_loop = m_pl->selectLoop();
        m_loopIdx = -1;
        if (m_loop->downloadable())
        {
            Downloader *d = new Downloader(m_loop);
            connect(d, SIGNAL(downloadDone()), this, SLOT(showLoopNext()));
            d->start();
            return;
        }
        else
            m_loopIdx = 0; /* no download, no event, so continue */
    }

    PlaylistItem *item = m_loop->at(m_loopIdx);

    qDebug() << item->uri() << item->type() << item->href();

    if (item->href())
    {
        showLoopNext();
        return;
    }

    m_prevWidget = m_currentWidget;

    switch (item->type())
    {
        case PlaylistItem::TypeIdle:
            m_currentWidget = new Idle(item, m_viewer);
            break;
        case PlaylistItem::TypeImage:
            m_currentWidget = new Image(item, m_viewer);
            break;
        case PlaylistItem::TypeVideo:
            m_currentWidget = new Video(item, m_viewer);
            break;
        case PlaylistItem::TypePdf:
            m_currentWidget = new PDF(item, m_viewer);
            break;
        case PlaylistItem::TypeWebpage:
            m_currentWidget = new Webpage(item, m_viewer);
            break;
        case PlaylistItem::TypeDocument:
            m_currentWidget = new Document(item, m_viewer);
            break;
        default:
            next();
            return;
    }

    if ((dynamic_cast<Idle *>(m_prevWidget) != NULL)
        && (item->type() != PlaylistItem::TypeIdle))
    {
        qDebug() << "Stopping idle";
        QProcess p;
        p.start("xset dpms force on");
        p.waitForFinished();
    }

    m_viewer->next(dynamic_cast<QWidget *>(m_currentWidget));
    if (m_prevWidget)
        dynamic_cast<QWidget *>(m_prevWidget)->deleteLater();
}
